from trypnv import Machine, PluginStateMachine
from trypnv.nvim import HasNvim

from proteome.nvim import NvimFacade
from proteome.logging import Logging

from tryp import List


class ProteomeComponent(Machine, HasNvim, Logging):

    def __init__(self, name: str, vim: NvimFacade) -> None:
        Machine.__init__(self, name)
        HasNvim.__init__(self, vim)


class ProteomeState(PluginStateMachine, HasNvim, Logging):

    def __init__(self, vim: NvimFacade, plugins: List[str]) -> None:
        self.vim = vim
        PluginStateMachine.__init__(self, 'proteome', plugins)
        HasNvim.__init__(self, vim)


__all__ = ['ProteomeComponent', 'ProteomeState']



from pathlib import Path

from proteome.project import Project
from proteome.nvim import NvimFacade

from trypnv import ProcessExecutor, Job  # type: ignore


class Git(ProcessExecutor):

    def pre_args(self, project: Project):
        return []

    def command(self, project: Project, name: str, *cmd_args):
        args = self.pre_args(project) + [name] + list(cmd_args)
        self.log.debug('running git {}'.format(' '.join(args)))
        return self.run(Job(project, 'git', args, self.loop))

    # TODO remove dangling lock file
    # and set the excludesfile
    def init(self, project: Project):
        return self.command(project, 'init')

    def add_commit_all(self, project: Project, message: str):
        return self.command(project, 'add', '-A', str(project.root))\
            .flat_map(lambda a: self.command(project, 'commit', '-m', message))


class HistoryGit(Git):

    def __init__(self, base: Path, vim: NvimFacade, loop=None) -> None:
        self.base = base
        super(HistoryGit, self).__init__(vim, loop)

    def pre_args(self, project: Project):
        d = str(project.root)
        h = str(self._history_dir(project))
        return [
            '--git-dir',
            h,
            '--work-tree',
            d,
        ]

    def _history_dir(self, project: Project):
        return self.base / project.fqn

    @property
    def ready(self):
        return self.base.is_dir() and super(HistoryGit, self).ready

__all__ = ['Git']



from pathlib import Path
from typing import Callable, Any
from datetime import datetime

from tryp import _
from tryp.lazy import lazy

from trypnv.machine import may_handle, message

from proteome.state import ProteomeComponent
from proteome.env import Env
from proteome.git import HistoryGit
from proteome.project import Project
from proteome.plugins.core import Save, StageIV

Commit = message('Commit')
HistoryBufferPrev = message('HistoryBufferPrev')
HistoryBufferNext = message('HistoryBufferNext')


class Plugin(ProteomeComponent):

    @lazy
    def git(self):
        base = self.vim.pdir('history_base').get_or_else(Path('/dev/null'))
        return HistoryGit(base, self.vim)

    # TODO handle broken repo
    def _commit(self, pro: Project):
        self.log.debug('commiting to history repo for {}'.format(pro))
        return self.git.add_commit_all(pro, datetime.now().isoformat())

    def _init(self, pro: Project):
        self.log.debug('initializing history repo for {}'.format(pro))
        return self.git.init(pro)

    @property
    def all_projects_history(self):
        return self.pflags.all_projects_history

    @property
    def git_ready(self):
        return self.git is not None and self.git.ready

    def _handle(self, env: Env, handler: Callable[[Project], Any]):
        name = handler.__name__
        if self.git_ready:
            projects = env.all_projects
            if not self.all_projects_history:
                projects = projects.filter(_.history)
            inf = 'running history handler {} on {}'
            self.log.verbose(inf.format(name, projects))
            projects.map(handler)
            self.git.exec()
        else:
            err = 'tried to run {} on history while not ready'
            self.log.verbose(err.format(name))

    @may_handle(StageIV)
    def stage_4(self, env: Env, msg):
        self._handle(env, self._init)

    @may_handle(Commit)
    def commit(self, env: Env, msg):
        self._handle(env, self._commit)

    @may_handle(Save)
    def save(self, env, msg):
        return Commit()

    @may_handle(HistoryBufferPrev)
    def history_buffer_prev(self, env, msg):
        pass

__all__ = ['Commit', 'Plugin']



from typing import Callable, Any
from contextlib import contextmanager

from flexmock import flexmock  # type: ignore

import tryp
import tryp.test
from tryp import may, Maybe
from tryp.logging import tryp_stdout_logging

from proteome.nvim import NvimFacade

from trypnv.nvim import Buffer


class MockNvimFacade(NvimFacade):

    def __init__(self):
        self.vars = {}
        super(MockNvimFacade, self).__init__(None)
        self.target = self

    @may
    def var(self, name: str) -> Maybe[str]:  # type: ignore
        v = self.vars.get(name)
        if v is None:
            self.log.error('variable not found: {}'.format(name))
        return v

    @property
    def current_buffer(self):
        return Buffer(self, self, self.prefix)

    def switch_root(self, root):
        pass

    def async(self, f: Callable[['NvimFacade'], Any]):
        return f(self)

    @contextmanager
    def main_event_loop(self):
        yield

    def cmd(self, *a, **kw):
        pass


class Spec(tryp.test.Spec):

    def setup(self, *a, **kw):
        tryp.development = True
        tryp_stdout_logging()
        super(Spec, self).setup(*a, **kw)


class MockNvimSpec(Spec):

    def setup(self, *a, **kw):
        super(MockNvimSpec, self).setup(*a, **kw)
        self.vim = MockNvimFacade()
        self.vim_mock = flexmock(self.vim)

__all__ = ['Spec']
